/*
 *  grppasswd utils
 * 
 *  Copyright (C) 1997  Ragnar Hojland Espinosa
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *  You may contact the author for ideas, comments or bug reports at
 * 
 *                                    tech.support@redestb.es
 *                                    ragnar@dfyre.ccon.net
 */


#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <pwd.h>
#include <errno.h>
#include <stdarg.h>
#include <pwd.h>
#include <getopt.h>
#include <stdlib.h>
#include "common.h"

char rootpasswd[0xff];

struct magic_uid_entry
  {
    const char * const name;
    const uid_t uid;
    bool seen;
  };

#include "pwchk.h"


int main (int argc, char *argv[])
  {
    int o, longindex;
    bool success;

    const struct option longopts[] =
      {
        { "help",        no_argument, 0, 'h' },
	{ "version",     no_argument, 0, 'V' },
	{ "verbose",     optional_argument, 0, 'v' },
	{ "strict",      optional_argument, 0, 's' },
	{ "interactive", optional_argument, 0, 'i' },
	{ 0, 0, 0, 0 }
      };

    flag.verbose     = DEFAULT_VERBOSE;
    flag.strict      = DEFAULT_STRICT;
    flag.interactive = DEFAULT_INTERACTIVE;
    
    while ( (o=getopt_long (argc, argv, "vsihV", longopts, &longindex)) != -1)
    switch (o)
      {
	case 'h':
	  showhelp();
	  return 0;
	
	case 'V': 
	  showversion();
	  return 0;
	
	case 'v':
	  if (optarg)
	    {
	      flag.verbose = atoi (optarg);
	      break;
	    }	
	  flag.verbose++;
	  break;
	
	case 's':
	  if (optarg)
	    {
	      flag.strict = atoi (optarg);
	      break;
	    }
	  flag.strict++;
	  break;
	
	case 'i':
	  if (optarg)
	    {
	      flag.interactive = atoi (optarg);
	      break;
	    }
	  flag.interactive++;
	  break;
	
	default:
	  fprintf (stderr, "Try `%s --help' for more information.\n",
		   argv[0]);
   	  return 1;
      }
    
    
    success = checkpasswdfile();
    if (success) 
      {
	verbose (2, "passwd file check: passed\n");
	return 0;
      }
    else
      {
	verbose (1, "passwd file check: failed\n");
	return 1;
      }

    return 1;
  }

void showhelp (void)
  {
    printf ("Usage: pwchk [OPTION]\n\n"
            "  -v, --verbose=NUM       increments verbosity\n"
	    "  -s, --strict=NUM        increments strictness\n"
	    "  -i, --interactive=NUM   increments interactivity\n"
	    "  -h, --help              displays this help\n"
	    "  -V  --version           displays version information\n"
            "\n"
            "-vvv is equivalent to --verbose=3\n");
  }

void showversion (void)
  {
    printf (PWCHK_VERSION_INFO
	    "Compiled %s:  verbose %d  strict %d  interactive %d\n",
	    __DATE__, DEFAULT_VERBOSE, DEFAULT_STRICT, DEFAULT_INTERACTIVE);
  }

void verbose (int level, char *format, ...)
  {
    if (level <= flag.verbose)
      {
	char buffer[1024];
	va_list (args);
	va_start (args, format);
	vsprintf (buffer, format, args);
	va_end (args);
	
	printf (buffer);
      }
  }

bool checkpasswdfile (void)
  {
    struct passwd *pw;
    
    setpwent();

    while ( (pw = getpwent()) )
      {
	if (!checkpwent (pw))
	  {
	    return false;
	  }
      }
    
    endpwent();
    
    if (flag.interactive && rootpasswd)
      {
	struct magic_uid_entry *e = findUidEntry ("root");
	char salt[2];
	char buffer[0xff];
	
	if (!e)
	  {
	    verbose (1, "no root\n");
	    return false;
	  }

	strcpy (buffer, getpass ("root's password: "));
       	salt[0] = rootpasswd[0];
	salt[1] = rootpasswd[1];

	if (strcmp (crypt(buffer,salt), rootpasswd))
	  {
	    verbose (1, "invalid password\n");
	    return false;
	  }
	else
	  {
	    verbose (2, "valid password\n");
	  }
      }
    
    return checkmagicuids();
  }


bool checkmagicuids (void)
  {
    int i;
    
    for (i = 0; magic_uid[i].name; i++)
      {
	if (!magic_uid[i].seen)
	  {
	    verbose (1, "magic_uid missing: %s (%d)\n", 
		     magic_uid[i].name, magic_uid[i].uid);
	    return false;
	  }
      }
    
    return true;
  }
   

struct magic_uid_entry* findUidEntry (const char * const name)
  {
    int i = 0;

    if (name[0] == '\0')  { return 0; }
    
    for (i = 0; magic_uid[i].name; i++)
      {
	if (!strcmp (magic_uid[i].name, name))
	  {
	    return &magic_uid[i];
	  }
      }
    
    return 0;
  }


bool checkpwent (struct passwd *pw)
  {
    bool name_ok   = pw->pw_name[0];
    bool passwd_ok = pw->pw_passwd[0];
    bool dir_ok    = pw->pw_dir[0];
    bool shell_ok  = pw->pw_shell[0];

    struct magic_uid_entry *e;
    struct stat stat_buf;
    
    if (!name_ok)
      {
	verbose (1, "name missing\n");
	return false;
      }

    if (!passwd_ok)
      {
	if ( (flag.strict) || (pw->pw_uid == 0 || pw->pw_gid == 0))
	  {
	    verbose (1, "password missing: user %s (%d/%d)\n", 
		     pw->pw_name, pw->pw_uid, pw->pw_gid);
	    return false;
	  }
      }

    if (!strcmp ("root", pw->pw_name))
      {
	strncpy (rootpasswd, pw->pw_passwd, sizeof(rootpasswd));
      }
    
    if (!dir_ok)
      {
	if (flag.strict)
	  {
	    verbose (1, "directory missing: user %s (%d/%d)\n",
		     pw->pw_name, pw->pw_uid, pw->pw_gid);
	    return false;
	  }
      }
    else
      {
	const int i = stat (pw->pw_dir, &stat_buf);
	
	if (i == -1)
	  {
	    if (errno == ENOENT)
	      {
		verbose (1, "can't find directory: %s\n", 
			 pw->pw_dir);
		return false;
	      }
	    else
	      {
		verbose (1, "error looking for directory: %s\n",
			 pw->pw_dir);
		return false;
	      }
	  }
	else
	  {
	    // we dont test explicitely for S_ISDIR in case we have /dev/null
	    if (S_ISREG(stat_buf.st_mode))
	      {
		verbose (1, "directory is a regular file: %s\n", pw->pw_dir);
		return false;
	      }
	    else if (S_ISDIR(stat_buf.st_mode))
	      {
		if (!(stat_buf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)))
		  {
		    verbose (1, "directory isn't executable by anybody: %s\n",
			     pw->pw_dir);
		    return false;
		  }
	      }
	    
	    if (pw->pw_dir[0] != '/')
	      {
		verbose (1, "directory name isn't absolute: %s\n",
			 pw->pw_dir);
		return false;
		
	      }
	  }
	
      }
    
    if (!shell_ok)
      {
	if (flag.strict)
	  {
	    verbose (1, "shell missing: user %s (%d/%d)\n", 
		     pw->pw_name, pw->pw_uid, pw->pw_gid);
	    return false;
	  }
      }
    else
      {
	const int i = pw->pw_shell ?
	        stat (pw->pw_shell, &stat_buf) : stat ("/bin/sh", &stat_buf);
	
	if (i == -1)
	  {
	    if (errno == ENOENT)
	      {
		verbose (1, "can't find shell: %s\n",
			 pw->pw_shell);
		return false;
	      }
	    else
	      {
		verbose (1, "error looking for shell\n",
			 pw->pw_shell);
		return false;
	      }
	  }
	else
	  {
	    const bool isroot = !strcmp ("root", pw->pw_name);
	    if (!S_ISREG(stat_buf.st_mode))
	      {
		if (isroot)
		  {
		    verbose (1, "root's shell isn't a regular file: %s\n",
			     pw->pw_shell);
		    return false;
		  }
	      }
	    else 
	      {
		if (!(stat_buf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)))
		  {
		    verbose (1, "shell isn't executable by anybody: %s\n",
			     pw->pw_shell);
		    return false;
		  }

		if (pw->pw_shell[0] != '/')
		  {
		    verbose (1, "shell name isn't absolute: %s\n",
			     pw->pw_shell);
		    return false;
		
		  }	      
	      }
	    
	    if (flag.strict > 1)
	      {
		char buffer[8192];
		FILE *fp = fopen ("/etc/shells", "r");
		bool found = false;
		
		if (!fp)
		  {
		    verbose (1, "couldn't open /etc/shells\n");
		    return false;
		  }
		
		do
		  {
		    fgets (buffer, sizeof(buffer), fp);
		    if (!strcmp (buffer, pw->pw_shell))
		      {
			found = true;
			break;
		      }
		  }
		while (!feof(fp));
		fclose (fp);
		
		if (!found)
		  {
		    verbose (1, "couldn't find %s in /etc/shells\n",
			     pw->pw_shell);
		    return false;
		  }
	      }
	  }
	

      }
    
    e = findUidEntry (pw->pw_name);
    if (e)
      {
	if (e->uid != pw->pw_uid)
	  {
	    verbose (1, "unmatching uid: %s is %d, should be %d\n",
		     pw->pw_name, pw->pw_uid, e->uid);
	    return false;
	  }
	
	if (e->seen)
	  {
	    verbose (1, "duplicated uid: %s (%d)\n",
		     pw->pw_name, pw->pw_uid);
	    return false;
	  }
	
	e->seen = true;
      }

    return true;
  }
